import os
from collections import OrderedDict

import yaml


OPTION_DEFAULTS = {
    "type": "str",
    "unknown_option": False,
    "default": None,
    "desc": None,
}


def ordered_load(stream):
    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return OrderedDict(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    OrderedLoader.add_constructor('!include', OrderedLoader.include)

    return yaml.load(stream, OrderedLoader)


class OrderedLoader(yaml.Loader):
    # This class was pulled out of ordered_load() for the sake of
    # mocking __init__ in a unit test.
    def __init__(self, stream):
        self._root = os.path.split(stream.name)[0]
        super(OrderedLoader, self).__init__(stream)

    def include(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, 'r') as f:
            return yaml.load(f, OrderedLoader)


def ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
    class OrderedDumper(Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            list(data.items()))
    OrderedDumper.add_representer(OrderedDict, _dict_representer)
    return yaml.dump(data, stream, OrderedDumper, **kwds)



"""
Functionality for dealing with dbkeys.
"""
# dbkeys read from disk using builds.txt
from __future__ import absolute_import

import os.path
from json import loads

from galaxy.util import read_dbnames
from galaxy.util.object_wrapper import sanitize_lists_to_string


class GenomeBuilds(object):
    default_value = "?"
    default_name = "unspecified (?)"

    def __init__(self, app, data_table_name="__dbkeys__", load_old_style=True):
        self._app = app
        self._data_table_name = data_table_name
        self._static_chrom_info_path = app.config.len_file_path
        # A dbkey can be listed multiple times, but with different names, so we can't use dictionaries for lookups
        if load_old_style:
            self._static_dbkeys = list(read_dbnames(app.config.builds_file_path))
        else:
            self._static_dbkeys = []

    def get_genome_build_names(self, trans=None):
        # FIXME: how to deal with key duplicates?
        rval = []
        # load user custom genome builds
        if trans is not None:
            if trans.history:
                # This is a little bit Odd. We are adding every .len file in the current history to dbkey list,
                # but this is previous behavior from trans.db_names, so we'll continue to do it.
                # It does allow one-off, history specific dbkeys to be created by a user. But we are not filtering,
                # so a len file will be listed twice (as the build name and again as dataset name),
                # if custom dbkey creation/conversion occurred within the current history.
                datasets = trans.sa_session.query(self._app.model.HistoryDatasetAssociation) \
                                .filter_by(deleted=False, history_id=trans.history.id, extension="len")
                for dataset in datasets:
                    rval.append((dataset.dbkey, "%s (%s) [History]" % (dataset.name, dataset.dbkey)))
            user = trans.user
            if user and hasattr(user, 'preferences') and 'dbkeys' in user.preferences:
                user_keys = loads(user.preferences['dbkeys'])
                for key, chrom_dict in user_keys.items():
                    rval.append((key, "%s (%s) [Custom]" % (chrom_dict['name'], key)))
        # Load old builds.txt static keys
        rval.extend(self._static_dbkeys)
        # load dbkeys from dbkey data table
        dbkey_table = self._app.tool_data_tables.get(self._data_table_name, None)
        if dbkey_table is not None:
            for field_dict in dbkey_table.get_named_fields_list():
                rval.append((field_dict['value'], field_dict['name']))
        return rval

    def get_chrom_info(self, dbkey, trans=None, custom_build_hack_get_len_from_fasta_conversion=True):
        # FIXME: flag to turn off custom_build_hack_get_len_from_fasta_conversion should not be required
        chrom_info = None
        db_dataset = None
        # Collect chromInfo from custom builds
        if trans:
            db_dataset = trans.db_dataset_for(dbkey)
            if db_dataset:
                chrom_info = db_dataset.file_name
            else:
                # Do Custom Build handling
                if trans.user and ('dbkeys' in trans.user.preferences) and (dbkey in loads(trans.user.preferences['dbkeys'])):
                    custom_build_dict = loads(trans.user.preferences['dbkeys'])[dbkey]
                    # HACK: the attempt to get chrom_info below will trigger the
                    # fasta-to-len converter if the dataset is not available or,
                    # which will in turn create a recursive loop when
                    # running the fasta-to-len tool. So, use a hack in the second
                    # condition below to avoid getting chrom_info when running the
                    # fasta-to-len converter.
                    if 'fasta' in custom_build_dict and custom_build_hack_get_len_from_fasta_conversion:
                        # Build is defined by fasta; get len file, which is obtained from converting fasta.
                        build_fasta_dataset = trans.sa_session.query(trans.app.model.HistoryDatasetAssociation).get(custom_build_dict['fasta'])
                        chrom_info = build_fasta_dataset.get_converted_dataset(trans, 'len').file_name
                    elif 'len' in custom_build_dict:
                        # Build is defined by len file, so use it.
                        chrom_info = trans.sa_session.query(trans.app.model.HistoryDatasetAssociation).get(custom_build_dict['len']).file_name
        # Check Data table
        if not chrom_info:
            dbkey_table = self._app.tool_data_tables.get(self._data_table_name, None)
            if dbkey_table is not None:
                chrom_info = dbkey_table.get_entry('value', dbkey, 'len_path', default=None)
        # use configured server len path
        if not chrom_info:
            # Default to built-in build.
            # Since we are using an unverified dbkey, we will sanitize the dbkey before use
            chrom_info = os.path.join(self._static_chrom_info_path, "%s.len" % sanitize_lists_to_string(dbkey))
        chrom_info = os.path.abspath(chrom_info)
        return (chrom_info, db_dataset)



"""Entry point for the usage of Cheetah templating within Galaxy."""
from __future__ import absolute_import

import sys
import traceback
from lib2to3.refactor import RefactoringTool

import packaging.version
from Cheetah.Compiler import Compiler
from Cheetah.NameMapper import NotFound
from Cheetah.Template import Template
from past.translation import myfixes

from . import unicodify

# Skip libpasteurize fixers, which make sure code is py2 and py3 compatible.
# This is not needed, we only translate code on py3.
if sys.version_info.major > 2:
    myfixes = [f for f in myfixes if not f.startswith('libpasteurize')]
    refactoring_tool = RefactoringTool(myfixes, {'print_function': True})
else:
    myfixes = refactoring_tool = None


class FixedModuleCodeCompiler(Compiler):

    module_code = None

    def getModuleCode(self):
        self._moduleDef = self.module_code
        return self._moduleDef


def create_compiler_class(module_code):

    class CustomCompilerClass(FixedModuleCodeCompiler):
        pass

    setattr(CustomCompilerClass, 'module_code', module_code)

    return CustomCompilerClass


def fill_template(template_text,
                  context=None,
                  retry=10,
                  compiler_class=Compiler,
                  first_exception=None,
                  futurized=False,
                  python_template_version='3',
                  **kwargs):
    """Fill a cheetah template out for specified context.

    If template_text is None, an exception will be thrown, if context
    is None (the default) - keyword arguments to this function will be used
    as the context.
    """
    if template_text is None:
        raise TypeError("Template text specified as None to fill_template.")
    if not context:
        context = kwargs
    if isinstance(python_template_version, str):
        python_template_version = packaging.version.parse(python_template_version)
    klass = Template.compile(source=template_text, compilerClass=compiler_class)
    t = klass(searchList=[context])
    try:
        return unicodify(t)
    except NotFound as e:
        if first_exception is None:
            first_exception = e
        if refactoring_tool and python_template_version.release[0] < 3 and retry > 0:
            tb = e.__traceback__
            last_stack = traceback.extract_tb(tb)[-1]
            if last_stack.name == '<listcomp>':
                # On python 3 list, dict and set comprehensions as well as generator expressions
                # have their own local scope, which prevents accessing frame variables in cheetah.
                # We can work around this by replacing `$var` with `var`, but we only do this for
                # list comprehensions, as this has never worked for dict or set comprehensions or
                # generator expressions in Cheetah.
                var_not_found = e.args[0].split("'")[1]
                replace_str = 'VFFSL(SL,"%s",True)' % var_not_found
                lineno = last_stack.lineno - 1
                module_code = t._CHEETAH_generatedModuleCode.splitlines()
                module_code[lineno] = module_code[lineno].replace(replace_str, var_not_found)
                module_code = "\n".join(module_code)
                compiler_class = create_compiler_class(module_code)
                return fill_template(template_text=template_text,
                                     context=context,
                                     retry=retry - 1,
                                     compiler_class=compiler_class,
                                     first_exception=first_exception,
                                     python_template_version=python_template_version,
                                     )
        raise first_exception or e
    except Exception as e:
        if first_exception is None:
            first_exception = e
        if sys.version_info.major > 2 and python_template_version.release[0] < 3 and not futurized:
            # Possibly an error caused by attempting to run python 2
            # template code on python 3. Run the generated module code
            # through futurize and hope for the best.
            module_code = t._CHEETAH_generatedModuleCode
            module_code = futurize_preprocessor(module_code)
            compiler_class = create_compiler_class(module_code)
            return fill_template(template_text=template_text,
                                 context=context,
                                 retry=retry,
                                 compiler_class=compiler_class,
                                 first_exception=first_exception,
                                 futurized=True,
                                 python_template_version=python_template_version,
                                 )
        raise first_exception or e


def futurize_preprocessor(source):
    source = str(refactoring_tool.refactor_string(source, name='auto_translate_cheetah'))
    # libfuturize.fixes.fix_unicode_keep_u' breaks from Cheetah.compat import unicode
    source = source.replace('from Cheetah.compat import str', 'from Cheetah.compat import unicode')
    return source



"""
A simple wrapper for writing tarballs as a stream.
"""
from __future__ import absolute_import

import logging
import os
import tarfile

from galaxy.exceptions import ObjectNotFound
from .path import safe_walk

log = logging.getLogger(__name__)


class StreamBall(object):
    def __init__(self, mode, members=None):
        self.members = members
        if members is None:
            self.members = []
        self.mode = mode
        self.wsgi_status = None
        self.wsgi_headeritems = None

    def add(self, file, relpath, check_file=False):
        if check_file and len(file) > 0:
            if not os.path.isfile(file):
                raise ObjectNotFound
            else:
                self.members.append((file, relpath))
        else:
            self.members.append((file, relpath))

    def stream(self, environ, start_response):
        response_write = start_response(self.wsgi_status, self.wsgi_headeritems)

        class tarfileobj(object):
            def write(self, *args, **kwargs):
                response_write(*args, **kwargs)
        tf = tarfile.open(mode=self.mode, fileobj=tarfileobj())
        for (file, rel) in self.members:
            tf.add(file, arcname=rel)
        tf.close()
        return []


class ZipBall(object):
    def __init__(self, tmpf, tmpd):
        self._tmpf = tmpf
        self._tmpd = tmpd

    def stream(self, environ, start_response):
        response_write = start_response(self.wsgi_status, self.wsgi_headeritems)
        tmpfh = open(self._tmpf)
        response_write(tmpfh.read())
        tmpfh.close()
        try:
            os.unlink(self._tmpf)
            os.rmdir(self._tmpd)
        except OSError:
            log.exception("Unable to remove temporary library download archive and directory")
        return []


def stream_archive(trans, path, upstream_gzip=False):
    archive_type_string = 'w|gz'
    archive_ext = 'tgz'
    if upstream_gzip:
        archive_type_string = 'w|'
        archive_ext = 'tar'
    archive = StreamBall(mode=archive_type_string)
    for root, directories, files in safe_walk(path):
        for filename in files:
            p = os.path.join(root, filename)
            relpath = os.path.relpath(p, os.path.join(path, os.pardir))
            archive.add(file=os.path.join(path, p), relpath=relpath)
    archive_name = "%s.%s" % (os.path.basename(path), archive_ext)
    trans.response.set_content_type("application/x-tar")
    trans.response.headers["Content-Disposition"] = 'attachment; filename="{}"'.format(archive_name)
    archive.wsgi_status = trans.response.wsgi_status()
    archive.wsgi_headeritems = trans.response.wsgi_headeritems()
    return archive.stream
